@@ -11,7 +11,7 @@
 #'
 #' @docType class
 CFAxis <- R6::R6Class("CFAxis",
-  inherit = CFObject,
+  inherit = CFData,
   cloneable = FALSE,
   private = list(
     # A character "X", "Y", "Z" or "T" to indicate the
@@ -47,7 +47,7 @@ CFAxis <- R6::R6Class("CFAxis",
     # extend this method.
     get_coordinates = function() {
       crds <- if (private$.active_coords == 1L) {
-        private$.values
+        self$values
       } else
         private$.aux[[private$.active_coords - 1L]]$coordinates
       dim(crds) <- NULL
@@ -58,7 +58,7 @@ CFAxis <- R6::R6Class("CFAxis",
     # ax will receive the auxiliary coordinates subsetted by argument rng.
     subset_coordinates = function(ax, rng) {
       if (length(private$.aux)) {
-        lapply(private$.aux, function(x) ax$auxiliary <- x$subset(x$name, rng))
+        lapply(private$.aux, function(x) ax$auxiliary <- x$subset(x$name, group = ax$group, rng))
       }
     },
 
@@ -71,7 +71,7 @@ CFAxis <- R6::R6Class("CFAxis",
       ax$unlimited <- private$.unlimited
 
       if (inherits(private$.bounds, "CFBounds"))
-        ax$bounds <- private$.bounds$copy()
+        ax$bounds <- private$.bounds$subset(group = ax$group, rng)
 
       private$subset_coordinates(ax, rng)
       ax$active_coordinates <- self$active_coordinates
@@ -88,6 +88,7 @@ CFAxis <- R6::R6Class("CFAxis",
     #'   Creating a new axis is more easily done with the [makeAxis()] function.
     #' @param var The name of the axis when creating a new axis. When reading an
     #'   axis from file, the [NCVariable] object that describes this instance.
+    #' @param group The [CFGroup] that this instance will live in.
     #' @param values Optional. The values of the axis in a vector. Ignored when
     #'   argument `var` is a `NCVariable` object.
     #' @param start Optional. Integer index where to start reading axis data
@@ -101,15 +102,19 @@ CFAxis <- R6::R6Class("CFAxis",
     #'   `NCVariable` instance, attributes of the axis will be taken from the
     #'   netCDF resource.
     #' @return A basic `CFAxis` object.
-    initialize = function(var, values, start = 1L, count = NA, orientation = "", attributes = data.frame()) {
+    initialize = function(var, group, values, start = 1L, count = NA, orientation = "", attributes = data.frame()) {
       orientation <- orientation[1L]
       if (!(orientation %in% c("X", "Y", "Z", "T", "")))
         stop("Invalid orientation for the axis.", call. = FALSE) # nocov
 
-      super$initialize(var, values = values, start = start, count = count, attributes = attributes)
+      # Scalar axes don't have .NC_map, so catch NULL values from $copy() etc
+      if (is.null(start)) start <- 1L
+      if (is.null(count)) count <- NA
+
+      super$initialize(var, group = group, values = values, start = start, count = count, attributes = attributes)
       if (inherits(var, "NCVariable")) {
-        private$.unlimited <- if (private$.NCvar$ndims) private$.NCvar$dimension(1L)$unlim else FALSE
-        private$.dimid <- private$.NCvar$dimids[1L]
+        private$.unlimited <- if (private$.NCobj$ndims) private$.NCobj$dimension(1L)$unlim else FALSE
+        private$.dimid <- private$.NCobj$dimids[1L]
       }
 
       private$.orient <- orientation
@@ -165,8 +170,12 @@ CFAxis <- R6::R6Class("CFAxis",
     #' @return Character string with very basic axis information.
     shard = function() {
       unlim <- if (self$unlimited) "-U" else ""
-      crds <- self$coordinate_range
-      paste0("[", self$name, " (", self$length, unlim, "): ", crds[1L], " ... ", crds[2L], "]")
+      if (self$length == 1L)
+        paste0("[", self$name, " (", self$length, unlim, "): ", self$coordinates, "]")
+      else {
+        crds <- self$coordinate_range
+        paste0("[", self$name, " (", self$length, unlim, "): ", crds[1L], " ... ", crds[2L], "]")
+      }
     },
 
     #' @description Retrieve interesting details of the axis.
@@ -222,14 +231,15 @@ CFAxis <- R6::R6Class("CFAxis",
     },
 
     #' @description Tests if the axis passed to this method is identical to
-    #'   `self`. This only tests for generic properties - class, length and name - with
-    #'   further assessment done in sub-classes.
+    #'   `self`. This only tests for generic properties - class, length, name
+    #'   and attributes - with further assessment done in sub-classes.
     #' @param axis The `CFAxis` instance to test.
     #' @return `TRUE` if the two axes are identical, `FALSE` if not.
     identical = function(axis) {
-      all(class(self) == class(axis)) &&
+      class(self)[1L] == class(axis)[1L] &&
       self$length == axis$length &&
-      self$name == axis$name
+      self$name == axis$name &&
+      self$attributes_identical(axis$attributes)
     },
 
     #' @description Tests if the axis passed to this method can be appended to
@@ -240,7 +250,7 @@ CFAxis <- R6::R6Class("CFAxis",
     #'   not.
     can_append = function(axis) {
       all(class(self) == class(axis)) &&
-      mode(private$.values) == mode(axis$values) &&
+      mode(self$values) == mode(axis$values) &&
       self$name == axis$name
     },
 
@@ -250,8 +260,9 @@ CFAxis <- R6::R6Class("CFAxis",
     #'   `CFAxis` descendants that do not implement this method.
     #' @param name The name for the new axis. If an empty string is passed, will
     #'   use the name of this axis.
+    #' @param group The [CFGroup] where the copy of this axis will live.
     #' @return `NULL`
-    copy = function(name = "") {
+    copy = function(name = "", group) {
       NULL
     },
 
@@ -262,9 +273,10 @@ CFAxis <- R6::R6Class("CFAxis",
     #'   implement this method.
     #' @param name The name for the new axis. If an empty string is passed, will
     #'   use the name of this axis.
+    #' @param group The [CFGroup] where the copy of this axis will live.
     #' @param values The values to the used with the copy of this axis.
     #' @return `NULL`
-    copy_with_values = function(name = "", values) {
+    copy_with_values = function(name = "", group, values) {
       NULL
     },
 
@@ -274,11 +286,12 @@ CFAxis <- R6::R6Class("CFAxis",
     #'   succeed with no result for the  `CFAxis` descendants that do not
     #'   implement this method.
     #' @param name The name for the new axis if the `rng` argument is provided.
+    #' @param group The [CFGroup] where the copy of this axis will live.
     #' @param rng The range of indices whose values from this axis to include in
     #'   the returned axis. If the value of the argument is `NULL`, return a
     #'   copy of the axis.
     #' @return `NULL`
-    subset = function(name = "", rng = NULL) {
+    subset = function(name = "", group, rng = NULL) {
       NULL
     },
 
@@ -300,40 +313,83 @@ CFAxis <- R6::R6Class("CFAxis",
       stop("`indexOf()` must be implemented by descendant CFAxis class.")
     },
 
-    #' @description Write the axis to a netCDF file, including its attributes.
-    #' @param nc The handle of the netCDF file opened for writing or a group in
-    #'   the netCDF file. If `NULL`, write to the file or group where the axis
-    #'   was read from (the file must have been opened for writing).
+    #' @description Attach this axis to a group. If there is another object with
+    #'   the same name in this group an error is thrown. For associated objects
+    #'   (such as bounds, etc), if another object with the same name is
+    #'   otherwise identical to the associated object then that object will be
+    #'   linked from the variable, otherwise an error is thrown.
+    #' @param grp An instance of [CFGroup].
+    #' @param locations Optional. A `list` whose named elements correspond to
+    #'   the names of objects associated with this axis, possibly including the
+    #'   axis itself. Each list element has a single character string indicating
+    #'   the group in the hierarchy where the object should be stored. As an
+    #'   example, if the variable has axes "lon" and "lat" and they should be
+    #'   stored in the parent group of `grp`, then specify `locations = list(lon
+    #'   = "..", lat = "..")`. Locations can use absolute paths or relative
+    #'   paths from group `grp`. The axis and associated objects that are not in
+    #'   the list will be stored in group `grp`. If the argument `locations` is
+    #'   not provided, all associated objects will be stored in this group.
     #' @return Self, invisibly.
-    write = function(nc = NULL) {
-      h <- if (inherits(nc, "NetCDF")) nc else private$.NCvar$handle
+    attach_to_group = function(grp, locations = list()) {
+      # Boundary values and auxiliary coordinates first
+      if (!is.null(private$.bounds))
+        private$.bounds$attach_to_group(grp, locations)
+      lapply(private$.aux, function(aux) aux$attach_to_group(grp, locations))
 
-      private$.id <- if (self$length == 1L && !private$.unlimited)
-        RNetCDF::var.def.nc(h, self$name, private$.data_type, NA)
-      else {
-        # Write the dimension for the axis. Error will be thrown when trying to
-        # write a dimension that's already defined, such as when a dimension is
-        # shared between multiple objects. In that case, get the id.
-        did <- try(if (private$.unlimited)
-          RNetCDF::dim.def.nc(h, self$name, unlim = TRUE)
-          else
-            RNetCDF::dim.def.nc(h, self$name, self$length), silent = TRUE)
-        if (inherits(did, "try-error"))
-          did <- RNetCDF::dim.inq.nc(h, self$name)$id
+      # Now add the axis to the group
+      super$attach_to_group(grp, locations)
+    },
 
-        RNetCDF::var.def.nc(h, self$name, private$.data_type, did)
+    #' @description Write the axis to a netCDF file, including its attributes.
+    #' @return Self, invisibly.
+    write = function() {
+      if (is.null(private$.NCobj)) {
+        # Try to find a NC variable and its associated dimension
+        ncobj <- private$.group$NC$find_by_name(self$name)
+        if (is.null(ncobj)) {
+          # Create a new NC dimension and variable on file.
+          if (self$length == 1L && !private$.unlimited)
+            private$.NCobj <- NCVariable$new(id = NA, name = self$name, group = private$.group$NC,
+                                             vtype = private$.data_type, dimids = NA)
+          else {
+            private$.dimid <- NCDimension$new(id = NA, name = self$name, length = self$length,
+                                              unlim = private$.unlimited, group = private$.group$NC)$id
+            private$.NCobj <- NCVariable$new(id = NA, name = self$name, group = private$.group$NC,
+                                             vtype = private$.data_type, dimids = private$.dimid)
+          }
+        } else if (inherits(ncobj, "NCDimension")) {
+          # NC dimension was found so NC variable does not exist
+          if (self$length == ncobj$length && private$.unlimited == ncobj$unlim) {
+            private$.dimid <- ncobj$id
+            private$.NCobj <- NCVariable$new(id = NA, name = self$name, group = private$.group$NC,
+                                             vtype = private$.data_type, dimids = private$.dimid)
+          } else
+            stop("Incompatible dimension by this name already exists.", call. = FALSE)
+        } else {
+          # NC variable was found
+          if (private$.data_type == ncobj$vtype && ncobj$ndims == 1L &&
+              self$length == ncobj$dimension(1L)$length &&
+              private$.unlimited == ncobj$dimension(1L)$unlim) {
+            private$.NCobj <- ncobj
+            private$.dimid <- ncobj$dimids
+          } else
+            stop("Incompatible object by this name already exists.", call. = FALSE)
+        }
+        private$.id <- private$.NCobj$id
+        private$.data_dirty <- TRUE
+        private$write_data()
       }
 
       if (private$.orient %in% c("X", "Y", "Z", "T"))
         self$set_attribute("axis", "NC_CHAR", private$.orient)
-      self$write_attributes(h, self$name)
-
-      RNetCDF::var.put.nc(h, self$name, private$.values)
+      self$write_attributes()
 
+      # Bounds
       if (!is.null(private$.bounds))
-        private$.bounds$write(h, self$name)
+        private$.bounds$write(self$name)
 
-      lapply(private$.aux, function(l) l$write(nc, did))
+      # Auxiliary coordinates
+      lapply(private$.aux, function(l) {l$dimid <- private$.dimid; l$write()})
 
       invisible(self)
     }
@@ -357,7 +413,7 @@ CFAxis <- R6::R6Class("CFAxis",
     #' @field length (read-only) The declared length of this axis.
     length = function(value) {
       if (missing(value))
-        length(private$.values)
+        length(self$values)
     },
 
     #' @field orientation Set or retrieve the orientation of the axis, a single
@@ -397,7 +453,7 @@ CFAxis <- R6::R6Class("CFAxis",
       if (missing(value))
         private$.bounds
       else if (inherits(value, "CFBounds")) {
-        # FIXME: Check the bounds values
+        # FIXME: Check the bounds values, could be climatology
         private$.bounds <- value
         self$set_attribute("bounds", "NC_CHAR", value$name)
       } else if (!is.null(value))
@@ -414,7 +470,7 @@ CFAxis <- R6::R6Class("CFAxis",
         if (private$.active_coords == 1L) NULL
         else private$.aux[[private$.active_coords - 1L]]
       } else {
-        if ((inherits(value, "CFLabel") || inherits(value, "CFAxis")) &&
+        if (inherits(value, c("CFLabel", "CFAxis")) &&
             value$length == self$length && !(value$name %in% private$aux_names())) {
           private$.aux <- append(private$.aux, setNames(list(value), value$name))
         }
@@ -424,8 +480,8 @@ CFAxis <- R6::R6Class("CFAxis",
     #' @field coordinate_names (read-only) Retrieve the names of the coordinate
     #'   sets defined for the axis, as a character vector. The first element in
     #'   the vector is the name of the axis and it refers to the values of the
-    #'   coordinates as stored in the netCDF file. Following elements refer to
-    #'   auxiliary coordinates.
+    #'   coordinates of this axis. Following elements refer to auxiliary
+    #'   coordinates.
     coordinate_names = function(value) {
       if (missing(value))
         c(self$name, names(private$.aux))
@@ -455,14 +511,14 @@ CFAxis <- R6::R6Class("CFAxis",
       }
     },
 
-    #' @field unlimited Logical to indicate if the axis is unlimited.
+    #' @field unlimited Logical to indicate if the axis is unlimited. The
+    #'   setting can only be changed if the axis has not yet been wriiten to
+    #'   file.
     unlimited = function(value) {
       if (missing(value))
         private$.unlimited
-      else if (is.logical(value[1L])) {
+      else if (is.null(private$.NCobj) && is.logical(value))
         private$.unlimited <- value[1L]
-      } else
-        stop("Must use single logical value to set or unset unlimited axis.", call. = FALSE) # nocov
     },
 
     #' @field time (read-only) Retrieve the `CFTime` object associated with the
@@ -488,14 +544,17 @@ CFAxis <- R6::R6Class("CFAxis",
 #'
 #' This method returns the lengths of the axes of a variable or axis.
 #'
-#' @param x The `CFVariable` or a descendant of `CFAxis`.
-#' @return Vector of axis lengths.
+#' @param x A `CFVariable` instance or a descendant of `CFAxis`.
+#' @return For a `CFVariable` instance in argument `x`, a named vector of axis
+#'   lengths, excluding any scalar axes. For a `CFAxis` descendant instance in
+#'   argument `x`, the length of the axis.
 #' @export
 #' @examples
 #' fn <- system.file("extdata", "ERA5land_Rwanda_20160101.nc", package = "ncdfCF")
 #' ds <- open_ncdf(fn)
 #' t2m <- ds[["t2m"]]
 #' dim(t2m)
+#' dim(t2m$axes[["time"]])
 dim.CFAxis <- function(x) {
   x$length
 }
@@ -509,6 +568,7 @@ dimnames.CFAxis <- function(x) {
 #' @param object A `CFAxis` instance or any descendant.
 #' @param ... Ignored.
 #' @export
+#' @keywords internal
 str.CFAxis <- function(object, ...) {
   cat(paste0("<", object$friendlyClassName, "> ", object$shard(), "\n"))
 }
