@@ -17,6 +17,7 @@ CFAxisLatitude <- R6::R6Class("CFAxisLatitude",
     #'   [makeLatitudeAxis()] function.
     #' @param var The name of the axis when creating a new axis. When reading an
     #'   axis from file, the [NCVariable] object that describes this instance.
+    #' @param group The [CFGroup] that this instance will live in.
     #' @param values Optional. The values of the axis in a vector. The values
     #'   have to be numeric within the range (-90, 90) and monotonic. Ignored
     #'   when argument `var` is a NCVariable object.
@@ -28,11 +29,10 @@ CFAxisLatitude <- R6::R6Class("CFAxisLatitude",
     #'   axis. When an empty `data.frame` (default) and argument `var` is an
     #'   NCVariable instance, attributes of the axis will be taken from the
     #'   netCDF resource.
-    initialize = function(var, values, start = 1L, count = NA, attributes = data.frame()) {
-      super$initialize(var, values = values, start = start, count = count, orientation =  "Y", attributes = attributes)
+    initialize = function(var, group, values, start = 1L, count = NA, attributes = data.frame()) {
+      super$initialize(var, group, values = values, start = start, count = count, orientation =  "Y", attributes = attributes)
       self$set_attribute("standard_name", "NC_CHAR", "latitude")
-      if (is.na(self$attribute("units")))
-        self$set_attribute("units", "NC_CHAR", "degrees_north")
+      self$set_attribute("units", "NC_CHAR", "degrees_north")
 
       if (!missing(values) && !.check_latitude_domain(values))
         stop("Latitude coordinate values are not in the domain [-90, 90].", call. = FALSE) # nocov
@@ -43,17 +43,18 @@ CFAxisLatitude <- R6::R6Class("CFAxisLatitude",
     #' from new instances.
     #' @param name The name for the new axis. If an empty string is passed, will
     #'   use the name of this axis.
+    #' @param group The [CFGroup] where the copy of this axis will live.
     #' @return The newly created axis.
-    copy = function(name = "") {
+    copy = function(name = "", group) {
       if (self$has_resource) {
-        ax <- CFAxisLatitude$new(self$NCvar, values = private$.values, start = private$.start_count$start,
-                                 count = private$.start_count$count, attributes = self$attributes)
+        ax <- CFAxisLatitude$new(self$NC, group = group, values = self$values, start = private$.NC_map$start,
+                                 count = private$.NC_map$count, attributes = self$attributes)
         if (nzchar(name))
           ax$name <- name
       } else {
         if (!nzchar(name))
           name <- self$name
-        ax <- CFAxisLatitude$new(name, values = private$.values, attributes = self$attributes)
+        ax <- CFAxisLatitude$new(name, group = group, values = self$values, attributes = self$attributes)
       }
       private$copy_properties_into(ax)
     },
@@ -67,39 +68,41 @@ CFAxisLatitude <- R6::R6Class("CFAxisLatitude",
     #'   should set, modify or delete attributes as appropriate.
     #' @param name The name for the new axis. If an empty string is passed, will
     #'   use the name of this axis.
+    #' @param group The [CFGroup] where the copy of this axis will live.
     #' @param values The values to the used with the copy of this axis.
     #' @return The newly created axis.
-    copy_with_values = function(name = "", values) {
+    copy_with_values = function(name = "", group, values) {
       if (!nzchar(name))
         name <- self$name
-      CFAxisLatitude$new(name, values = values, attributes = self$attributes)
+      CFAxisLatitude$new(name, group = group, values = values, attributes = self$attributes)
     },
 
-    #' @description Return alatitude axis spanning a smaller coordinate range.
+    #' @description Return a latitude axis spanning a smaller coordinate range.
     #'   This method returns an axis which spans the range of indices given by
     #'   the `rng` argument.
     #' @param name The name for the new axis. If an empty string is passed
     #'   (default), will use the name of this axis.
+    #' @param group The [CFGroup] where the copy of this axis will live.
     #' @param rng The range of indices whose values from this axis to include in
     #'   the returned axis. If the value of the argument is `NULL`, return a
     #'   copy of the axis.
     #' @return A new `CFAxisLatitude` instance covering the indicated range of
     #'   indices. If the value of the argument `rng` is `NULL`, return a copy of
     #'   `self` as the new axis.
-    subset = function(name = "", rng = NULL) {
+    subset = function(name = "", group, rng = NULL) {
       if (is.null(rng))
-        self$copy(name)
+        self$copy(name, group)
       else {
         rng <- range(rng)
         if (self$has_resource) {
-          ax <- CFAxisLatitude$new(private$.NCvar, start = private$.start_count$start + rng[1L] - 1L,
+          ax <- CFAxisLatitude$new(private$.NCobj, group = group, start = private$.NC_map$start + rng[1L] - 1L,
                                    count = rng[2L] - rng[1L] + 1L, attributes = self$attributes)
           if (nzchar(name))
             ax$name <- name
         } else {
           if (!nzchar(name))
             name <- self$name
-          ax <- CFAxisLatitude$new(name, values = private$.values[rng[1L]:rng[2L]], attributes = self$attributes)
+          ax <- CFAxisLatitude$new(name, group = group, values = self$values[rng[1L]:rng[2L]], attributes = self$attributes)
         }
         private$copy_properties_into(ax, rng)
       }
@@ -111,15 +114,16 @@ CFAxisLatitude <- R6::R6Class("CFAxisLatitude",
     #'   resulting axis.
     #' @param from An instance of `CFAxisLatitude` whose values to append to the
     #'   values of this axis.
+    #' @param group The [CFGroup] where the copy of this axis will live.
     #' @return A new `CFAxisLatitude` instance with values from this axis and
     #'   the `from` axis appended.
     append = function(from) {
       if (super$can_append(from) && .c_is_monotonic(self$values, from$values)) {
-        ax <- CFAxisLatitude$new(self$name, values = c(private$values, from$values),
+        ax <- CFAxisLatitude$new(self$name, group = group, values = c(private$values, from$values),
                                  attributes = self$attributes)
 
         if (!is.null(private$.bounds)) {
-          new_bnds <- private$.bounds$append(from$bounds)
+          new_bnds <- private$.bounds$append(from$bounds, group)
           if (!is.null(new_bnds))
             ax$bounds <- new_bnds
         }
